@@ -1,46 +1,60 @@
+from typing import Union
+
 import numpy as np
 
 from kwave.executor import Executor
 from kwave.kWaveSimulation import kWaveSimulation
 from kwave.kWaveSimulation_helper import retract_transducer_grid_size, save_to_disk_func
-from kwave.kspaceFirstOrder import kspaceFirstOrderC, kspaceFirstOrderG
+from kwave.kgrid import kWaveGrid
+from kwave.kmedium import kWaveMedium
+from kwave.ksensor import kSensor
+from kwave.ksource import kSource
+from kwave.ktransducer import NotATransducer
+from kwave.options.simulation_execution_options import SimulationExecutionOptions
+from kwave.options.simulation_options import SimulationOptions
 from kwave.utils.dotdictionary import dotdict
 from kwave.utils.interp import interpolate3d
 from kwave.utils.pml import get_pml
 from kwave.utils.tictoc import TicToc
 
 
-@kspaceFirstOrderG
-def kspaceFirstOrder3DG(**kwargs):
+def kspaceFirstOrder3DG(
+        kgrid: kWaveGrid,
+        source: kSource,
+        sensor: Union[NotATransducer, kSensor],
+        medium: kWaveMedium,
+        simulation_options: SimulationOptions,
+        execution_options: SimulationExecutionOptions
+):
     """
-        3D time-domain simulation of wave propagation on a GPU using C++ CUDA code.
+    3D time-domain simulation of wave propagation on a GPU using C++ CUDA code.
 
-        kspaceFirstOrder3DG provides a blind interface to the C++/CUDA
-        version of kspaceFirstOrder3D (called kspaceFirstOrder-CUDA) in the
-        same way as kspaceFirstOrder3DC. Note, the C++ code does not support
-        all input options, and all display options are ignored (only command
-        line outputs are given). See the k-Wave user manual for more
-        information.
+    kspaceFirstOrder3DG provides a blind interface to the C++/CUDA
+    version of kspaceFirstOrder3D (called kspaceFirstOrder-CUDA) in the
+    same way as kspaceFirstOrder3DC. Note, the C++ code does not support
+    all input options, and all display options are ignored (only command
+    line outputs are given). See the k-Wave user manual for more
+    information.
 
-        The function works by appending the optional input 'save_to_disk' to
-        the user inputs and then calling kspaceFirstOrder3D to save the input
-        files to disk. The contents of sensor.record (if set) are parsed as
-        input flags, and the C++ code is run using the system command. The
-        output files are then automatically loaded from disk and returned in
-        the same fashion as kspaceFirstOrder3D. The input and output files
-        are saved to the temporary directory native to the operating system,
-        and are deleted after the function runs.
+    The function works by appending the optional input 'save_to_disk' to
+    the user inputs and then calling kspaceFirstOrder3D to save the input
+    files to disk. The contents of sensor.record (if set) are parsed as
+    input flags, and the C++ code is run using the system command. The
+    output files are then automatically loaded from disk and returned in
+    the same fashion as kspaceFirstOrder3D. The input and output files
+    are saved to the temporary directory native to the operating system,
+    and are deleted after the function runs.
 
-        This function requires the C++ binary/executable of
-        kspaceFirstOrder-CUDA to be downloaded from
-        http://www.k-wave.org/download.php and placed in the "binaries"
-        directory of the k-Wave toolbox. Alternatively, the name and location
-        of the binary can be specified using the optional input parameters
-        'BinaryName' and 'BinariesPath'.
+    This function requires the C++ binary/executable of
+    kspaceFirstOrder-CUDA to be downloaded from
+    http://www.k-wave.org/download.php and placed in the "binaries"
+    directory of the k-Wave toolbox. Alternatively, the name and location
+    of the binary can be specified using the optional input parameters
+    'BinaryName' and 'BinariesPath'.
 
-        This function is essentially a wrapper and directly uses the capabilities
-        of kspaceFirstOrder3DC by replacing the binary name with the name of the
-        GPU binary.
+    This function is essentially a wrapper and directly uses the capabilities
+    of kspaceFirstOrder3DC by replacing the binary name with the name of the
+    GPU binary.
 
     Args:
         **kwargs:
@@ -48,42 +62,57 @@ def kspaceFirstOrder3DG(**kwargs):
     Returns:
 
     """
-    sensor_data = kspaceFirstOrder3DC(**kwargs)  # pass inputs to CPU version
+    execution_options.is_gpu_simulation = True
+    assert execution_options.is_gpu_simulation, 'kspaceFirstOrder2DG can only be used for GPU simulations'
+    sensor_data = kspaceFirstOrder3D(
+        kgrid=kgrid,
+        source=source,
+        sensor=sensor,
+        medium=medium,
+        simulation_options=simulation_options,
+        execution_options=execution_options
+    )  # pass inputs to CPU version
     return sensor_data
 
 
-@kspaceFirstOrderC()
-def kspaceFirstOrder3DC(**kwargs):
+def kspaceFirstOrder3DC(
+        kgrid: kWaveGrid,
+        source: kSource,
+        sensor: Union[NotATransducer, kSensor],
+        medium: kWaveMedium,
+        simulation_options: SimulationOptions,
+        execution_options: SimulationExecutionOptions
+):
     """
-        3D time-domain simulation of wave propagation using C++ code.
+    3D time-domain simulation of wave propagation using C++ code.
 
-        kspaceFirstOrder3DC provides a blind interface to the C++ version of
-        kspaceFirstOrder3D (called kspaceFirstOrder-OMP). Note, the C++ code
-        does not support all input options, and all display options are
-        ignored (only command line outputs are given). See the k-Wave user
-        manual for more information.
+    kspaceFirstOrder3DC provides a blind interface to the C++ version of
+    kspaceFirstOrder3D (called kspaceFirstOrder-OMP). Note, the C++ code
+    does not support all input options, and all display options are
+    ignored (only command line outputs are given). See the k-Wave user
+    manual for more information.
 
-        The function works by appending the optional input 'save_to_disk' to
-        the user inputs and then calling kspaceFirstOrder3D to save the input
-        files to disk. The contents of sensor.record (if set) are parsed as
-        input flags, and the C++ code is run using the system command. The
-        output files are then automatically loaded from disk and returned in
-        the same fashion as kspaceFirstOrder3D. The input and output files
-        are saved to the temporary directory native to the operating system,
-        and are deleted after the function runs.
+    The function works by appending the optional input 'save_to_disk' to
+    the user inputs and then calling kspaceFirstOrder3D to save the input
+    files to disk. The contents of sensor.record (if set) are parsed as
+    input flags, and the C++ code is run using the system command. The
+    output files are then automatically loaded from disk and returned in
+    the same fashion as kspaceFirstOrder3D. The input and output files
+    are saved to the temporary directory native to the operating system,
+    and are deleted after the function runs.
 
-        This function is not recommended for large simulations, as the input
-        variables will reside twice in main memory (once in MATLAB, and once
-        in C++). For large simulations, the C++ code should be called outside
-        of MATLAB. See the k-Wave manual for more information.
+    This function is not recommended for large simulations, as the input
+    variables will reside twice in main memory (once in MATLAB, and once
+    in C++). For large simulations, the C++ code should be called outside
+    of MATLAB. See the k-Wave manual for more information.
 
-        This function requires the C++ binary/executable of
-        kspaceFirstOrder-OMP to be downloaded from
-        http://www.k-wave.org/download.php and placed in the "binaries"
-        directory of the k-Wave toolbox (the same binary is used for
-        simulations in 2D, 3D, and axisymmetric coordinates). Alternatively,
-        the name and  location of the binary can be specified using the
-        optional input parameters 'BinaryName' and 'BinariesPath'.
+    This function requires the C++ binary/executable of
+    kspaceFirstOrder-OMP to be downloaded from
+    http://www.k-wave.org/download.php and placed in the "binaries"
+    directory of the k-Wave toolbox (the same binary is used for
+    simulations in 2D, 3D, and axisymmetric coordinates). Alternatively,
+    the name and  location of the binary can be specified using the
+    optional input parameters 'BinaryName' and 'BinariesPath'.
 
     Args:
         **kwargs:
@@ -91,12 +120,27 @@ def kspaceFirstOrder3DC(**kwargs):
     Returns:
 
     """
+    execution_options.is_gpu_simulation = False
     # generate the input file and save to disk
-    sensor_data = kspaceFirstOrder3D(**kwargs)
+    sensor_data = kspaceFirstOrder3D(
+        kgrid=kgrid,
+        source=source,
+        sensor=sensor,
+        medium=medium,
+        simulation_options=simulation_options,
+        execution_options=execution_options
+    )
     return sensor_data
 
 
-def kspaceFirstOrder3D(kgrid, medium, source, sensor, **kwargs):
+def kspaceFirstOrder3D(
+        kgrid: kWaveGrid,
+        source: kSource,
+        sensor: Union[NotATransducer, kSensor],
+        medium: kWaveMedium,
+        simulation_options: SimulationOptions,
+        execution_options: SimulationExecutionOptions
+):
     """
     3D time-domain simulation of wave propagation.
 
@@ -254,7 +298,20 @@ def kspaceFirstOrder3D(kgrid, medium, source, sensor, **kwargs):
     # start the timer and store the start time
     TicToc.tic()
 
-    k_sim = kWaveSimulation(kgrid, medium, source, sensor, **kwargs)
+    # Currently we only support binary execution, meaning all simulations must be saved to disk.
+    if not simulation_options.save_to_disk:
+        if execution_options.is_gpu_simulation:
+            raise ValueError('GPU simulation requires saving to disk. Please set SimulationOptions.save_to_disk=True')
+        else:
+            raise ValueError('CPU simulation requires saving to disk. Please set SimulationOptions.save_to_disk=True')
+
+    k_sim = kWaveSimulation(
+        kgrid=kgrid,
+        source=source,
+        sensor=sensor,
+        medium=medium,
+        simulation_options=simulation_options
+    )
     k_sim.input_checking('kspaceFirstOrder3D')
 
     # =========================================================================
@@ -345,7 +402,8 @@ def kspaceFirstOrder3D(kgrid, medium, source, sensor, **kwargs):
     if options.use_kspace:
         k = k_sim.kgrid.k
         k_sim.kappa = np.fft.ifftshift(np.sinc(c_ref * k * dt / 2))
-        if (k_sim.source_p and k_sim.source.p_mode == 'additive') or ((k_sim.source_ux or k_sim.source_uy or k_sim.source_uz) and k_sim.source.u_mode == 'additive'):
+        if (k_sim.source_p and k_sim.source.p_mode == 'additive') or \
+                ((k_sim.source_ux or k_sim.source_uy or k_sim.source_uz) and k_sim.source.u_mode == 'additive'):
             k_sim.source_kappa = np.fft.ifftshift(np.cos(c_ref * k * dt / 2))
     else:
         k_sim.kappa          = 1
@@ -362,7 +420,7 @@ def kspaceFirstOrder3D(kgrid, medium, source, sensor, **kwargs):
         retract_size = [[options.pml_x_size, options.pml_y_size, options.pml_z_size]]
 
         # run subscript to save files to disk
-        save_to_disk_func(k_sim.kgrid, k_sim.medium, k_sim.source, k_sim.options,
+        save_to_disk_func(k_sim.kgrid, k_sim.medium, k_sim.source, k_sim.options, execution_options.auto_chunking,
                           dotdict({
                               'ddx_k_shift_pos': k_sim.ddx_k_shift_pos,
                               'ddx_k_shift_neg': k_sim.ddx_k_shift_neg,
@@ -398,8 +456,8 @@ def kspaceFirstOrder3D(kgrid, medium, source, sensor, **kwargs):
 
                               'transducer_source': k_sim.transducer_source,
                               'nonuniform_grid': k_sim.nonuniform_grid,
-                              'elastic_code': k_sim.elastic_code,
-                              'axisymmetric': k_sim.axisymmetric,
+                              'elastic_code': k_sim.options.simulation_type.is_elastic_simulation(),
+                              'axisymmetric': k_sim.options.simulation_type.is_axisymmetric(),
                               'cuboid_corners': k_sim.cuboid_corners,
                           }))
 
@@ -410,7 +468,8 @@ def kspaceFirstOrder3D(kgrid, medium, source, sensor, **kwargs):
         if options.save_to_disk_exit:
             return
 
-        executor = Executor(device='gpu')
+        executor = Executor(simulation_options=simulation_options, execution_options=execution_options)
+        executor_options = execution_options.get_options_string(sensor=k_sim.sensor)
         sensor_data = executor.run_simulation(k_sim.options.input_filename, k_sim.options.output_filename,
-                                              options='--p_raw')
-        return k_sim.sensor.combine_sensor_data(sensor_data)
+                                              options=executor_options)
+        return sensor_data
